/*Version Notes
  v16r2 - Removed PID library which was difficult to tune and 
  changed to a simple regulation if/then ladder logic.
*/

/*Project Details
  Adjustable (in flight) RPM regulator for brushless motor ESC with 
  two temperature probes and data logging capability.
  ©Mark Qvale, 2019
  Creative Commons Attribution-ShareAlike 3.0 License
  Build article - http://www.itsqv.com/QVM/index.php?title=DIY_-_RPM_and_Temperature_Data_Logger
  
  Arduino Board and Pins used-
  Board - Arduino Pro Mini, Atmega 328P, 5v, 16mHz
  Pin 2 - RPM rpm sensor pin
  Pin 8 - Passthrough/Regulated switch channel
  Pin 9 - ESC out
  Pin 3 - RX/FC throttle in
  Pin 4 - Relay trigger out

  Arduino board to Micro SD Card Shield Pins-
  Pin 10 - CS
  pin 11 - MOSI
  pin 12 - MISO
  pin 13 - SCK(CLK)
  
 Items needed for this build-
  1x Hobbywing RPM sensor for brushless motors
  2x 100k or 10k thermistors
  1x Nano or Pro-Mini board
  1x Micro SD card shield
  1x SPDT Signal relay, 5v
  2x R1 resistors (see below)
  2x 10nF (103) capacitors
 
 Temp sensor & resistors-
 Glass bead thermistor 100k NTC3950
   R1 = 100K ohms
 10k thermistor
   R1 = 10K ohms
 Pin A0 is Temp 1 input
 Pin A1 is Temp 2 input

 Acknowledgements-
 Tom Igoe for his data log example code
 Adrian Li for the RPM counter section found at https://www.locarbftw.com/using-an-arduino-to-read-rpm-from-the-hobbywing-rpm-sensor/
*/

#include <Servo.h>                                        // Servo control library - Servo (built-in) by Michael Margolis
#include <Filter.h>                                       // Filter library - MegunoLink by Number Eight Innovation
#include <PWM.hpp>                                        // PWM reader library by Kamran Ahmad
#include <SPI.h>                                          // Serial Peripheral Interface library by Arduino
#include <SD.h>                                           // SD Card library by Arduino/Sparkfun
#define INTERRUPT_PIN 0                                   // defines pin 2 (0) as RPM interrupt
Servo escout;                                             // create servo/esc object

/* adjustable variables for signal smoothing  and PID parameters */
int rpmRND = 25;                                          // rounding value for rpm signal (default 25)
int EFW = 5;                                              // RPM exponential filter weight (default 10)
int EFWB = 40;                                            // Throttle signal exponential filter weight (default 40)
int RST = 2000;                                           // Regulator sample time in ms (default 3000)
int avgN = 20;                                            // temperature average number (default 20)
int period = 200;                                         // data print period in ms (default 200)
float numpoles = 14;                                      // number of motors magnet poles

/* global variables */
const int chipSelect = 10;                                // data write pin number
unsigned long time_now = 0;                               // print timer variable
unsigned long reg_time = 0;                               // regulator timeer variable
volatile int interruptCount;                              // RPM count variable
unsigned long time = 0;                                   // initialize time variable
int csw = 0;                                              // switch state key 0
int nsw = 1;                                              // switch state key 1
int switchState;                                          // flag for switch state
int relay = 4;                                            // relay pin
int thrOut;                                               // regulated ESC throttle output variable
float thpwm;                                              // RX/FC throttle input variable 
float thrP = 0;                                           // throttle percent variable
float rpm = 0;
float rpmset = 0;
int tempPin0 = A0;                                        // temperature read/compute variables & values
int tempPin1 = A1;                                        //
int Vo0;                                                  //
int Vo1;                                                  //
float avgT1;                                              //
float avgT2;                                              //
float R1 = 10000;                                         //
float logR2, R2, logR3, R3, T1, T2;                       //
float c1 = 1.009249522e-03;                               //
float c2 = 2.378405444e-04;                               //
float c3 = 2.019202697e-07;                               //

PWM thrIn(3);                                             // setup pin 3 for throttle input using the PWM ibrary

/* Exponential filter setup */
ExponentialFilter<long> ADCFilter(EFW, 0);                // rpm smoothing filter
ExponentialFilter<long> ADCFilterB(EFWB, 0);              // throttle signal smoothing filter

void setup()
{
  Serial.begin(115200);                                   // start the serial monitor
  thrIn.begin(true);                                      // begin reading throttle input (3) 
  escout.attach(9);                                       // attach ESC object (9)
  pinMode(relay, OUTPUT);                                 // relay on/off switch pin (4)
  pinMode(8, INPUT);                                      // set pinmode on RX switch pin (8)
  pinMode(INTERRUPT_PIN, INPUT);                          // set pinmode on RPM pin (2)
  attachInterrupt(INTERRUPT_PIN,interruptFired, CHANGE);  // set-up interrupt count on RPM pin (2)

  Serial.print("Initializing SD card...");                // see if the card is present and can be initialized
  if (!SD.begin(chipSelect)) {                            //
    Serial.println("Card fail, or not present");          //
    while(1);                                             //
  }
  Serial.println("card initialized.");                    //
  File dataFile = SD.open("Data_Log.csv", FILE_WRITE);    // name and open the CSV file
  if (dataFile) {
    dataFile.println("Time(ms),Temp 1(F),Temp 2(F),RPM,Thr%,Thr-v,Switch");     // write the column headings to the csv file
    dataFile.close();                                     // close file
  }
}

void loop(){
  time = millis();                                        // time in ms
/* +++++++++++++++++++++++++++++++++++++++++++ Throttle Value & Percent ++++++++++++++++++++++++++ */  
  thpwm = thrIn.getValue();                               // capture throttle signal
  ADCFilterB.Filter(thpwm);                               // throttle signal filter
  thpwm = ADCFilterB.Current();                           //
  thrP = thpwm;                                           // copy throttle signal value
  thrP = constrain(thrP, 990, 2010);                      // constrain values for mapping
  thrP = map(thrP, 990, 2010, 0, 100);                    // map throttle values to percent throttle
/* +++++++++++++++++++++++++++++++++++++++++++ Temp 1 & 2 +++++++++++++++++++++++++++++++++++++++++ */
  Vo0 = analogRead(tempPin0);                               // read temp pin 0
  R2 = R1 * (1023.0 / (float)Vo0 - 1.0);
  logR2 = log(R2);
  T1 = (1.0 / (c1 + c2*logR2 + c3*logR2*logR2*logR2));
  T1 = T1 - 273.15;                                         // temp in Celsius
  T1 = (T1 * 9.0)/ 5.0 + 32.0;                              // temp converted to Farenheit
  for(int i = 0; i < avgN; ++i){
    avgT1 += T1;
  }
  avgT1 /= avgN;
  avgT1 = int(avgT1);
  
  Vo1 = analogRead(tempPin1);                               // read temp pin 0
  R3 = R1 * (1023.0 / (float)Vo1 - 1.0);
  logR3 = log(R3);
  T2 = (1.0 / (c1 + c2*logR3 + c3*logR3*logR3*logR3));
  T2 = T2 - 273.15;                                         // temp in Celsius
  T2 = (T2 * 9.0)/ 5.0 + 32.0;                              // temp converted to Farenheit
  for(int i = 0; i < avgN; ++i){
    avgT2 += T2;
  }
  avgT2 /= avgN;
  avgT2 = int(avgT2);
/* +++++++++++++++++++++++++++++++++++++ Switch to throttle Pass through ++++++++++++++++++++++++++ */
  if (pulseIn(8, HIGH) > 1600){                             // pass-through switch position
    digitalWrite(relay, LOW);                               // holds relay closed (NC relay)
    csw=0;                                                  // switch state key
    switchState = csw;
  }
/* +++++++++++++++++++++++++++++++++++++ Switch to throttle RPM Regulation ++++++++++++++++++++++++ */
  if(pulseIn(8, HIGH) >= 900 && pulseIn(8, HIGH) <= 1600){  // regulate RPM switch position
    digitalWrite(relay, HIGH);                              // opens throttle pass-through relay
    if(nsw != csw){                                         // checks switch state
      rpmset = rpm;                                         // creates rpm signal set-point for the PID controller
      thrOut = thpwm+1;                                     // PWM transfer value
      csw = nsw;                                            // switch state cancel
    }
    switchState = nsw;
    if(millis() >= reg_time + RST){                         // start regulator timer
      reg_time = millis();                                  // reset regulator timer
      if(rpm < rpmset-25){                                  // throttle out adjust if RPM too low
        thrOut = thrOut + 1;
      }
      if(rpm > rpmset+25){                                  // throttle out adjust if RPM too high
        thrOut = thrOut - 1;
      }
    }
    escout.writeMicroseconds(thrOut);                       // send regulated throttle value to the ESC
  }
/* +++++++++++++++++++++++++++++++++++++ Switch Safety +++++++++++++++++++++++++++++++++++++++++++ */
  if(pulseIn(8, HIGH) < 900){                               // sets escout to 0 if no or low switch pulse detected (safety)
    escout.writeMicroseconds(0);
    digitalWrite(relay, LOW);                               // sets pass-through relay closed
  }
/* ++++++++++++++++++++++++++++++++++++++++++++ RPM Signal Capture +++++++++++++++++++++++++++++++ */
  noInterrupts();                                           // RPM count code
  interruptCount = 0;                                       //
  interrupts();                                             //
  delay(10);                                                //
  noInterrupts();                                           //
  int critical_rpm = interruptCount;                        //
  interrupts();                                             //
  rpm = ((critical_rpm)*(60))/(numpoles)*100;
  ADCFilter.Filter(rpm);                                    // exponential smoothing filter enable
  rpm = ADCFilter.Current();                                // smoothing filter output
  rpm = int(round(rpm/rpmRND))*rpmRND;                      // rpm converted to an integer and rounded
  if(rpm <= 1000){
    rpm = 0;
  }
/* ++++++++++++++++++++++++++++++++++++++++++++ Datafile Write ++++++++++++++++++++++++++++++++++ */
  if(millis() >= time_now + period){                        // start print timer
    time_now = millis();                                    // reset print timer
  
  File dataFile = SD.open("Data_Log.csv", FILE_WRITE);      // open the data log file
  if (dataFile) {                                           // if available, write
    dataFile.print(time);                                   // write to card - time in milli-seconds
    dataFile.print(", ");
    dataFile.print(avgT1);                                  // write to card - Temp 1
    dataFile.print(", ");
    dataFile.print(avgT2);                                  // write to card - Temp 2
    dataFile.print(", ");
    dataFile.print(rpm);                                    // write to card - RPM
    dataFile.print(", ");
    dataFile.print(thrP);                                   // write to card - Throttle percent
	  dataFile.print(", ");
    dataFile.print(thrOut);                                 // write to card - Throttle MS value
    dataFile.print(", ");
    dataFile.println(switchState);                          // write to card - Switch state
    dataFile.close();                                       // close data card file
  }
  else {
    Serial.println("Card error!");
  }
  }
/* ++++++++++++++++++++++++++++++++++++++++++++ Serial Monitor Output +++++++++++++++++++++++++++ */
  Serial.print("\n  Time: ");
  Serial.print(time);
  Serial.print("  Temp1: ");
  Serial.print(avgT1);
  Serial.print("  Temp2: ");
  Serial.print(avgT2);
  Serial.print("  RPM: ");
  Serial.print(rpm);
  Serial.print("  Throttle PWM: ");
  Serial.print(thrOut);
  Serial.print("  Switch POS: ");
  Serial.println(switchState);
}
/* ++++++++++++++++++++++++++++++++++++++++++++ RPM Counter Increment +++++++++++++++++++++++++++ */
void interruptFired(){
  interruptCount++;
}
